==> shader-ast-stdlib-0.12.27.tgz: package/color/porter-duff.js <==
import { defn, ret } from "@thi.ng/shader-ast/ast/function";
import { FLOAT0, FLOAT1, vec4 } from "@thi.ng/shader-ast/ast/lit";
import { add, mul, sub } from "@thi.ng/shader-ast/ast/ops";
import { $w } from "@thi.ng/shader-ast/ast/swizzle";
import { clamp01 } from "../math/clamp.js";
const coeff = (f, a, b) => (f === ZERO ? FLOAT0 : f === ONE ? a : mul(a, f($w(a), $w(b))));
/**
 * Higher-order Porter-Duff alpha compositing operator. See
 * [thi.ng/porter-duff](https://thi.ng/porter-duff) for reference. Returns an
 * optimized AST function which accepts 2 RGBA colors (vec4) and returns blended
 * & clamped result (also a vec4). All built-in PD operators are defined via
 * this HOF.
 *
 * The two given JS functions are used to extract blending coefficients for
 * src/dest colors and are called with the alpha components of both colors.
 *
 * Optimization only happens for cases where either `fa` and/or `fb` are
 * {@link ZERO}.
 *
 * @param name - function name
 * @param fa - src coeff fn
 * @param fb - dest coeff fn
 */
export const porterDuff = (name, fa, fb) => defn("vec4", name, ["vec4", "vec4"], (a, b) => {
    const src = coeff(fa, a, b);
    const dest = coeff(fb, a, b);
    const srcZero = src === FLOAT0;
    const destZero = dest === FLOAT0;
    return [
        ret(srcZero && destZero
            ? vec4()
            : clamp01(srcZero ? dest : destZero ? src : add(src, dest))),
    ];
});
// coefficient functions
export const ZERO = () => FLOAT0;
export const ONE = () => FLOAT1;
export const A = (a) => a;
export const B = (_, b) => b;
export const ONE_MINUS_A = (a) => sub(FLOAT1, a);
export const ONE_MINUS_B = (_, b) => sub(FLOAT1, b);
// standard Porter-Duff operators
export const blendSrcOver = porterDuff("blendSrcOver", ONE, ONE_MINUS_A);
export const blendDestOver = porterDuff("blendDestOver", ONE_MINUS_B, ONE);
export const blendSrcIn = porterDuff("blendSrcIn", B, ZERO);
export const blendDestIn = porterDuff("blendDestIn", ZERO, A);
export const blendSrcOut = porterDuff("blendSrcOut", ONE_MINUS_B, ZERO);
export const blendDestOut = porterDuff("blendDestOut", ZERO, ONE_MINUS_A);
export const blendSrcAtop = porterDuff("blendSrcAtop", B, ONE_MINUS_A);
export const blendDestAtop = porterDuff("blendDestAtop", ONE_MINUS_B, A);
export const blendXor = porterDuff("blendXor", ONE_MINUS_B, ONE_MINUS_A);
export const blendPlus = porterDuff("blendPlus", ONE, ONE);

==> shader-ast-stdlib-0.13.6.tgz: package/color/porter-duff.js <==
import { defn, ret } from "@thi.ng/shader-ast/ast/function";
import { FLOAT0, FLOAT1, vec4 } from "@thi.ng/shader-ast/ast/lit";
import { add, div, mul, sub } from "@thi.ng/shader-ast/ast/ops";
import { $w, $xyz } from "@thi.ng/shader-ast/ast/swizzle";
import { clamp01 } from "../math/clamp.js";
/** @internal */
const __coeff = (col, f) => f === FLOAT0 ? vec4() : f === FLOAT1 ? col : mul(col, f);
/**
 * Higher-order Porter-Duff alpha compositing operator. See
 * [thi.ng/porter-duff](https://thi.ng/porter-duff) for reference. Returns an
 * optimized AST function which accepts 2 RGBA colors (vec4) and returns blended
 * & clamped result (also a vec4).
 *
 * @remark
 * All built-in PD operators are defined via this HOF. The two given coefficient
 * functions are used to extract blending coefficients for src/dest colors and
 * are called with the alpha components of both colors.
 *
 * Optimization only happens for cases where either `fa` and/or `fb` are
 * {@link ZERO} or {@link ONE}.
 *
 * *IMPORTANT*: Both colors MUST be use pre-multiplied alpha for correct
 * results. If needed, use {@link premultiplyAlpha} and
 * {@link postmultiplyAlpha}.
 *
 * @param name - function name
 * @param fa - src coeff fn
 * @param fb - dest coeff fn
 */
export const porterDuff = (name, fa, fb) => defn("vec4", name, ["vec4", "vec4"], (a, b) => {
    const sa = fa($w(a), $w(b));
    const sb = fb($w(a), $w(b));
    const src = __coeff(a, sa);
    const dest = __coeff(b, sb);
    const srcZero = sa === FLOAT0;
    const destZero = sb === FLOAT0;
    return [
        ret(srcZero && destZero
            ? vec4()
            : clamp01(srcZero ? dest : destZero ? src : add(src, dest))),
    ];
});
export const premultiplyAlpha = (col) => vec4(mul($xyz(col), $w(col)), $w(col));
export const postmultiplyAlpha = (col) => vec4(div($xyz(col), $w(col)), $w(col));
// coefficient functions
export const ZERO = () => FLOAT0;
export const ONE = () => FLOAT1;
export const A = (a) => a;
export const B = (_, b) => b;
export const ONE_MINUS_A = (a) => sub(FLOAT1, a);
export const ONE_MINUS_B = (_, b) => sub(FLOAT1, b);
// standard Porter-Duff operators
export const blendSrcOver = porterDuff("blendSrcOver", ONE, ONE_MINUS_A);
export const blendDestOver = porterDuff("blendDestOver", ONE_MINUS_B, ONE);
export const blendSrcIn = porterDuff("blendSrcIn", B, ZERO);
export const blendDestIn = porterDuff("blendDestIn", ZERO, A);
export const blendSrcOut = porterDuff("blendSrcOut", ONE_MINUS_B, ZERO);
export const blendDestOut = porterDuff("blendDestOut", ZERO, ONE_MINUS_A);
export const blendSrcAtop = porterDuff("blendSrcAtop", B, ONE_MINUS_A);
export const blendDestAtop = porterDuff("blendDestAtop", ONE_MINUS_B, A);
export const blendXor = porterDuff("blendXor", ONE_MINUS_B, ONE_MINUS_A);
export const blendPlus = porterDuff("blendPlus", ONE, ONE);
